package JavaML;

import java.util.Arrays;

/**
 * An Online implementation of a two-class Binary stump
 *  
 * @author apocock
 */
public class OnlineBinaryClassifier implements IOnlineClassifier
{
  private int featureToUse;
  private boolean isOnePositive;
  
  private int[] totalCount;
  private int[] oneMapToOneCount;
  private int[] oneMapToNegOneCount;
  
  private boolean isTrained;
  
  public OnlineBinaryClassifier()
  {
    featureToUse = 0;
    isOnePositive = false;
  
    isTrained = false;
  }//default constructor

  /**
  * Copy constructor, used by the ensemble algorithms
  */
  public IOnlineClassifier copyClassifier()
  {
    IOnlineClassifier newBin = new OnlineBinaryClassifier();
    return newBin;
  }//copyClassifier
  
  /**
  * setupClassifier is used to modify classifiers after they have been constructed through the copy constructor,
  * at that point you could either do reflection to figure out the class and reconstruct it, or you could pass it
  * a string of arguments and have it reconfigure itself.
  *
  * This classifier has no parameters to modify
  *
  * @param  arguments   a space separated list of parameter=value tuples.
  */
  public void setupClassifier(String arguments)
  {
    String [] argsList = arguments.split(" ");
    try
    {
      for (int i = 0; i < argsList.length; i++)
      {
        String [] arg = argsList[i].split("=");
        if (arg.length == 2)
        {
        }//if arg formats
      }//for number of arguments
    }//try
    catch (NumberFormatException e)
    {
      System.out.println("Incorrectly formatted number in arguments");
    }
    catch (ArrayIndexOutOfBoundsException e)
    {
      System.out.println("Incorrectly formatted argument, missing equals");
    }
    catch (Exception e)
    {
      System.out.println("Other generic exception in setupClassifier");
      System.out.println(e.getMessage());
      e.printStackTrace(System.out);
    }
  }//setupClassifier(String)


  /**
  * Is the main offline train function, which takes an dataset and trains on it
  * 
  * in this classifier, iterations does nothing
  *
  * @param  trainingData    A training dataset
  * @param  iterations      Does nothing (there for IClassifier support for Boosting)
  */
  public double train(IDataset trainingData, int iterations)
  {
    double returnVal = 0.0;
        
    int numberOfSamples = trainingData.getNumberOfSamples();
    int numberOfFeatures = trainingData.getNumberOfFeatures();
    double [][] featureMatrix = trainingData.getDataMatrixDouble();
    int [] labels = trainingData.getTargetVector();
    double [] sampleWeights = trainingData.getDataWeights();
    int [] intWeights = new int[numberOfSamples];
    int totalWeight = trainingData.getTotalWeight();
    
    totalCount = new int[numberOfFeatures];
    oneMapToOneCount = new int[numberOfFeatures];
    oneMapToNegOneCount = new int[numberOfFeatures];
    
    double errorNormalise = 0.0;
    for (int i = 0; i < numberOfSamples; i++)
    {
      intWeights[i] = (int) Math.ceil(sampleWeights[i] * totalWeight);
      errorNormalise = errorNormalise + sampleWeights[i];
    }
    
    //System.out.println("intWeights[0] = " + intWeights[0]);
    //System.out.println("Error normalise = " + errorNormalise);
    
    
    double minError = errorNormalise;
    boolean isPosHigh = false;
    int selectedDimension = 0;
        
    //temp variables for the array
    double [] currentFeature;
    double threshold = 0.0;
    double tmpHighError = 0;
    double tmpLowError = 0;
    int highCorrect = 0;
    int lowCorrect = 0;

    //for each dimension
    for (int dim = 0; dim < numberOfFeatures; dim++)
    {
      currentFeature = featureMatrix[dim];
      tmpHighError = 0;
      tmpLowError = 0;
      highCorrect = 0;
      lowCorrect = 0;
        
    	//for each sample, test if 1 is +ve
    	for (int i = 0; i < numberOfSamples; i++)
    	{
        if (currentFeature[i] == 1.0)
        {
          if (labels[i] == 1)
          {
            tmpLowError += sampleWeights[i];
            highCorrect += intWeights[i];
          }
          else
          {
            tmpHighError += sampleWeights[i];
            lowCorrect += intWeights[i];
          }
        }
        else
        {
          if (labels[i] == 1)
          {
            tmpHighError += sampleWeights[i];
            lowCorrect += intWeights[i];
          }
          else
          {
            tmpLowError += sampleWeights[i];
            highCorrect += intWeights[i];
          }
        }
    	}//for number of samples
      
			oneMapToOneCount[dim] = highCorrect;
			oneMapToNegOneCount[dim] = lowCorrect;
    	totalCount[dim] = highCorrect + lowCorrect;
    	
    	if(tmpHighError < minError)
  		{
  			minError = tmpHighError;
  			isPosHigh = true;
  			selectedDimension = dim;
  		}
  		
  		if(tmpLowError < minError)
  		{
        minError = tmpLowError;
        isPosHigh = false;
  			selectedDimension = dim;
      }
    }//for number of features
    
    featureToUse = selectedDimension;
    //System.out.println("Feature to use = " + featureToUse);
    isOnePositive = isPosHigh;
    
    //System.out.println("PosHigh = " + isOnePositive);
       
    isTrained = true;
    
    //System.out.println("Selected dimension = " + selectedDimension + ", isOnePositive = " + isOnePositive + ", train error = " + minError/errorNormalise);
       
    return (minError/errorNormalise);
    //return returnVal;
  }//train(IDataset, int)
  
  
  /**
  * Is the online train function, which takes a single example and trains on it
  * then returns the classification.
  *
  * @param  trainingData    A training sample
  * @param  target          class label for the sample
  */
  public int onlineTrain(double[] trainingData, int target)
  {
    if (!isTrained)
    {
      totalCount = new int[trainingData.length];
      oneMapToOneCount = new int[trainingData.length];
      oneMapToNegOneCount = new int[trainingData.length];
    
      isTrained = true;
    }
    
    int newSelected = featureToUse;
    int largestDifference = 0;
    
    for (int i = 0; i < oneMapToOneCount.length; i++)
    {
      int currentValue = (int) trainingData[i];
      
      if ((currentValue == 1) && (target == 1))
      {
        oneMapToOneCount[i]++;
      }
      else if ((currentValue == 1) && (target == -1))
      {
        oneMapToNegOneCount[i]++;
      }
      else if ((currentValue == 0) && (target == 1))
      {
        oneMapToNegOneCount[i]++;
      }
      else if ((currentValue == 0) && (target == -1))
      {
        oneMapToOneCount[i]++;
      }
      totalCount[i]++;
      
      int temp = Math.abs(oneMapToNegOneCount[i] - oneMapToOneCount[i]);
      
      if (temp > largestDifference)
      {
        largestDifference = temp;
        newSelected = i;
      }
      //System.out.println("Started as " + isOnePositive);
    }//for number of features
    
    featureToUse = newSelected;
      
    if (oneMapToOneCount[featureToUse] > oneMapToNegOneCount[featureToUse])
    {
      isOnePositive = true;
      //System.out.println("Set positive");
    }
    else
    {
      isOnePositive = false;
      //System.out.println("Set negative");
    }
    
    return classifySample(trainingData);
  }//onlineTrain(double[],int)

  /**
  * Is the main offline test function, which takes an dataset and returns the testing error
  * on that dataset
  * 
  * @param  testingData    A testing dataset
  */
  public double test(IDataset testingData)
  {
    double returnVal = 0.0;
    if (isTrained)
    {
      //run test
      int numberOfTestSamples = testingData.getNumberOfSamples();
      int[] targetVector = testingData.getTargetVector();
      
      for (int i = 0; i < numberOfTestSamples; i++)
      {
        double[] currentSample = testingData.getSampleDouble(i,false);
        int finalClassification = classifySample(currentSample);
        if (finalClassification != targetVector[i])
        {
          returnVal = returnVal + 1.0;
        }
      }//for number of testing samples
      
      returnVal = returnVal / numberOfTestSamples;
    }//if trained
    else
    {
      returnVal = -1.0;
    }//else untrained, return -1
    return returnVal;
  }//test(IDataset)
  
  
  /**
  * Returns the classifications made by the trained classifier on the testing data
  * 
  * @param  testingData    A testing dataset
  */
  public int[] classifyDataset(IDataset testingData)
  {
    int[] classifications;
  
    if (isTrained)
    {
      int numberOfSamples = testingData.getNumberOfSamples();
      classifications = new int[numberOfSamples];
      double[][] featureMatrixPtr = testingData.getDataMatrixDouble();
      double[] featureToTest = featureMatrixPtr[featureToUse];
      
      for (int i = 0; i < numberOfSamples; i++)
      {
        if (featureToTest[i] == 1.0)
        {
          if (isOnePositive)
          {
            classifications[i] = IClassifier.positiveClass;
          }
          else
          {
            classifications[i] = IClassifier.negativeClass;
          }
        }
        else
        {
          if (!isOnePositive)
          {
            classifications[i] = IClassifier.positiveClass;
          }
          else
          {
            classifications[i] = IClassifier.negativeClass;
          }
        }
      }//for loop over number of samples
    }
    else
    {
      classifications = null;
    }
      
    return classifications;
  }//classifyDataset(IDataset)
        
  
  /**
  * Returns the classification for a single example
  * 
  * @param  sample        A testing sample
  */
  public int classifySample(double[] sample)
  {
    int classification;
  
    if (isTrained)
    {  
      if (sample[featureToUse] == 1.0)
      {
        if (isOnePositive)
        {
          classification = IClassifier.positiveClass;
        }
        else
        {
          classification = IClassifier.negativeClass;
        }
      }
      else
      {
        if (!isOnePositive)
        {
          classification = IClassifier.positiveClass;
        }
        else
        {
          classification = IClassifier.negativeClass;
        }
      }
    }
    else
    {
      classification = 0;
    }
      
    return classification;
  }//classifySample(double[])
  
  public int[] getFeatureList()
  {
    if (isTrained)
    {
      int [] returnArray = new int[1];
      
      returnArray[0] = featureToUse;
      
      return returnArray;
    }//if trained
    else
    {
      return new int[0];
    }//else untrained, return empty list
  }//getFeatureList()
  
  public boolean getIsTrained()
  {
    return isTrained;
  }
  
  public boolean supportsWeightedData()
  {
    return true;
  }
  
  public boolean supportsMultiClassData()
  {
    return false;
  }
}//public class ThresholdClassifier
